/*
 * lockable.hpp
 *
 * This file is part of Knights.
 *
 * Copyright (C) Stephen Thompson, 2006 - 2010.
 * Copyright (C) Kalle Marjola, 1994.
 *
 * Knights is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Knights is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Knights.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * Lockable: Base class for openable/trappable/lockable tiles (ie doors and chests).
 *
 */

#ifndef LOCKABLE_HPP
#define LOCKABLE_HPP

#include "tile.hpp"

#include "boost/shared_ptr.hpp"
using namespace boost;

class Creature;
class DungeonMap;
class MapCoord;
class Trap;

class Lockable : public Tile {
    enum { PICK_ONLY_LOCK_NUM=99998, SPECIAL_LOCK_NUM = 99999 };
    
public:
    Lockable() : closed(true), can_trap(true), lock(-1), lock_chance(0), pick_only_chance(0), keymax(1), trap_owner(0) { }

    void setOpenInitially() { closed = false; }
    void setLockChance(int lock_percent, int pick_only_percent, int kmax)
        { lock_chance = lock_percent; pick_only_chance = pick_only_percent; keymax = kmax; }
    void setCannotTrap() { can_trap = false; }
    void setSpecialLock() { lock = SPECIAL_LOCK_NUM; }

    // Override canActivateHere, onActivate, onOpen and onClose to deal with opening and
    // closing
    virtual bool canActivateHere() const { return false; }
    virtual void onActivate(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature>, Player *,
                            ActivateType, bool success = true);
    virtual void onClose(DungeonMap &, const MapCoord &, shared_ptr<Creature>, Player *);
    virtual void onOpen(DungeonMap &, const MapCoord &, shared_ptr<Creature>, Player *);
    
    bool isClosed() const { return closed; }
    
    // Locks
    bool isLocked() const { return lock > 0; }
    bool isSpecialLocked() const { return lock == SPECIAL_LOCK_NUM; }
    void generateLock(int nkeys);  // If "lock" is negative then this generates a lock.

    // return lock num (1,2,3 or PICK_ONLY_LOCK_NUM), or 0 (if not
    // locked, or special-locked). used by the dungeon generator for
    // connectivity checking.
    int getLockNum() const { return isLocked() && !isSpecialLocked() ? lock : 0; }

    
    // Traps
    // canTrap: true if this Lockable can have traps applied to it
    // setTrap: Set a trap which will be triggered when the door/chest is opened. If a trap
    // is already present, then the existing trap will be set off.
    // generateTrap: like setTrap but called during initial "pretrapped chests" generation.
    // (returns true if a trap was generated.)
    // onHit: Override this to deal with setting off traps when a door/chest is hit.
    bool canTrap() const { return can_trap; }
    void setTrap(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
                 shared_ptr<Trap> newtrap);
    virtual bool generateTrap(DungeonMap &, const MapCoord &) { return false; }
    virtual void onHit(DungeonMap &, const MapCoord &, shared_ptr<Creature>, Player *);

protected:
    // "open", "close" are implemented by subclasses, and handle the actual opening
    // and closing.
    virtual void open(DungeonMap &, const MapCoord &, Player *player) = 0;
    virtual void close(DungeonMap &, const MapCoord &, Player *player) = 0; 

private:
    bool doOpen(DungeonMap &, const MapCoord &, shared_ptr<Creature>, Player *, ActivateType);
    bool doClose(DungeonMap &, const MapCoord &, shared_ptr<Creature>, Player *, ActivateType);
    bool checkUnlock(shared_ptr<Creature>) const;
    void activateTraps(DungeonMap &, const MapCoord &, shared_ptr<Creature>);
    void disarmTraps(DungeonMap &, const MapCoord &, shared_ptr<Creature>);
    
private:
    bool closed;
    bool can_trap;
    int lock; // key number, or zero for unlocked
    // NOTE: a negative value for lock means that lock has not been generated yet.
    // See lock_chance and pick_only_chance for generation parameters.
    // ALSO: if lock is set to SPECIAL_LOCK_NUM then the door cannot be unlocked by keys
    // or lock picks; only switches will work.

    int lock_chance, pick_only_chance, keymax;
    shared_ptr<Trap> trap;
    Player *trap_owner;
};

#endif
